@font-face {
    font-family: 'NanumGothicCoding';
    src: url('../fonts/NanumGothicCoding-Regular.ttf') format('truetype');
    font-weight: normal;
    font-style: normal;
}

/****************************************************/

#workspace {
  position: absolute;
  width: 100%;
  height: 100%;
  display:flex;
  flex-direction: column;
  left: 0px;
  top: 0px;
  background-image: url("../img/background.jpg");
}

.view {
  overflow: auto;
  height: 200px;
  width: 400px;
  padding-left: 5px;
  padding-right: 5px;

  border-style: solid;
  border-width: 3px;
  border-radius: 5px;
  border-color: black;
  
  opacity: 0.9;
  color: black;
  background-color: white;

  font-family: 'NanumGothicCoding', serif;  
}

.view h1 {
  border-bottom: 1px solid black;
  font-size: 1em;
  background-color: rgb(230, 230, 230)  
}

.view p {
  text-align: left;
  font-size: 1.5em;
  white-space: pre;
}
